/****************************/
/* 8章 連立方程式の反復解法 */
/****************************/

/* 本書の例題8.2に対する連立ニュートン法（残差判定） */

#include<stdio.h>
#include<math.h>

#define N 2         // 行列の次元
#define EPS 1.0e-10 // 許容誤差
#define KMAX 20     // 最大反復回数
#define A 0.7       // 問題の係数A
#define B 0.2       // 問題の係数B

/* 関数値の計算 */
void function(double f[], double x[]){
  f[0] = x[0] - A*sin(x[0]) - B*cos(x[1]);
  f[1] = x[1] - A*cos(x[0]) - B*sin(x[1]);
  return;
}

/* ヤコビ行列の計算 */
void jacobian(double J[][N], double x[]){
  J[0][0] = 1.0 - A*cos(x[0]);
  J[0][1] = B*sin(x[1]);
  J[1][0] = A*sin(x[0]);
  J[1][1] = 1.0 - B*cos(x[1]);
  return;
}

/* ニュートン法の更新量（連立1次方程式の求解） */
void newton_step(double J[][N], double d[], double f[]){
  int i, j, k;
  double alpha, sum;

  /* Gaussの消去法（ピボット選択なし） */
  for(k=0; k<N-1; k++){
	for(i=k+1; i<N; i++){
	  alpha = J[i][k]/J[k][k];
	  for(j=k+1; j<N; j++){
		J[i][j] -= alpha*J[k][j];
	  }
	  f[i] -= alpha*f[k];
	}
  }
  for(i=N-1; i>=0; i--){
	sum = 0;
	for(j=i+1; j<N; j++){
	  sum += J[i][j]*d[j];
	}
	d[i] = (f[i] - sum)/J[i][i];
  }

  return;
}

/* ベクトルの2-ノルムを求める関数 */
double norm_2(double x[]){
  int i;
  double norm = 0.0;
  for(i=0; i<N; i++){
	norm += x[i]*x[i];
  }
  return sqrt(norm);
}

/* メイン関数 */
int main(void){
  int i, k;
  double J[N][N]; // ヤコビ行列
  double x[N];    // 近似解ベクトル
  double f[N];    // 残差ベクトル
  double d[N];    // ニュートン法の更新量を表すベクトル

  /* 初期近似解の設定 */
  for(i=0; i<N; i++){
	x[i] = 0.0;
  }
  
  printf("%2s: %18s: %18s\n", "k", "x_k", "y_k");
  for(i=0; i<42; i++){
	printf("-");
  }
  printf("\n");

  /* ニュートン反復 */
  k = 0;
  while(1){

	/* 近似解の出力 */
	printf("%2d: %.16f: %.16f\n", k, x[0], x[1]);
	
	/* 関数値（残差）の計算 */
	function(f, x);

	/* 収束判定 */
	if(norm_2(f) < EPS){
	  printf(" *** 収束しました ***\n");
	  break;
	}else if(k == KMAX){
	  printf(" *** 最大反復回数に到達 ***\n");
	  break;
	}
	
	/* ヤコビ行列の計算 */
	jacobian(J, x);

	/* 符号の反転 */
	for(i=0; i<N; i++){
	  f[i] = -f[i];
	}
	
	/* ニュートン法の更新量の計算 */
	newton_step(J, d, f);

	/* 近似解の更新 */
	for(i=0; i<N; i++){
	  x[i] += d[i];
	}

	/* 反復回数のカウント */
	k++;

  }

  return 0;
}
